# @generated by generate_proto_mypy_stubs.py.  Do not edit!
import sys
from google.protobuf.descriptor import (
    Descriptor as google___protobuf___descriptor___Descriptor,
)

from google.protobuf.internal.containers import (
    RepeatedScalarFieldContainer as google___protobuf___internal___containers___RepeatedScalarFieldContainer,
)

from google.protobuf.message import (
    Message as google___protobuf___message___Message,
)

from typing import (
    Iterable as typing___Iterable,
    Mapping as typing___Mapping,
    MutableMapping as typing___MutableMapping,
    Optional as typing___Optional,
    Text as typing___Text,
)

from typing_extensions import (
    Literal as typing_extensions___Literal,
)


builtin___bool = bool
builtin___bytes = bytes
builtin___float = float
builtin___int = int


class ErrorProto(google___protobuf___message___Message):
    DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...
    code = ... # type: builtin___int
    message = ... # type: typing___Text

    def __init__(self,
        *,
        code : typing___Optional[builtin___int] = None,
        message : typing___Optional[typing___Text] = None,
        ) -> None: ...
    @classmethod
    def FromString(cls, s: builtin___bytes) -> ErrorProto: ...
    def MergeFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    def CopyFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    if sys.version_info >= (3,):
        def ClearField(self, field_name: typing_extensions___Literal[u"code",u"message"]) -> None: ...
    else:
        def ClearField(self, field_name: typing_extensions___Literal[u"code",b"code",u"message",b"message"]) -> None: ...

class BranchRecord(google___protobuf___message___Message):
    DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...
    name = ... # type: typing___Text
    commit = ... # type: typing___Text

    def __init__(self,
        *,
        name : typing___Optional[typing___Text] = None,
        commit : typing___Optional[typing___Text] = None,
        ) -> None: ...
    @classmethod
    def FromString(cls, s: builtin___bytes) -> BranchRecord: ...
    def MergeFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    def CopyFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    if sys.version_info >= (3,):
        def ClearField(self, field_name: typing_extensions___Literal[u"commit",u"name"]) -> None: ...
    else:
        def ClearField(self, field_name: typing_extensions___Literal[u"commit",b"commit",u"name",b"name"]) -> None: ...

class HashRecord(google___protobuf___message___Message):
    DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...
    type = ... # type: typing___Text
    digest = ... # type: typing___Text

    def __init__(self,
        *,
        type : typing___Optional[typing___Text] = None,
        digest : typing___Optional[typing___Text] = None,
        ) -> None: ...
    @classmethod
    def FromString(cls, s: builtin___bytes) -> HashRecord: ...
    def MergeFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    def CopyFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    if sys.version_info >= (3,):
        def ClearField(self, field_name: typing_extensions___Literal[u"digest",u"type"]) -> None: ...
    else:
        def ClearField(self, field_name: typing_extensions___Literal[u"digest",b"digest",u"type",b"type"]) -> None: ...

class CommitRecord(google___protobuf___message___Message):
    DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...
    parent = ... # type: builtin___bytes
    ref = ... # type: builtin___bytes
    spec = ... # type: builtin___bytes

    def __init__(self,
        *,
        parent : typing___Optional[builtin___bytes] = None,
        ref : typing___Optional[builtin___bytes] = None,
        spec : typing___Optional[builtin___bytes] = None,
        ) -> None: ...
    @classmethod
    def FromString(cls, s: builtin___bytes) -> CommitRecord: ...
    def MergeFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    def CopyFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    if sys.version_info >= (3,):
        def ClearField(self, field_name: typing_extensions___Literal[u"parent",u"ref",u"spec"]) -> None: ...
    else:
        def ClearField(self, field_name: typing_extensions___Literal[u"parent",b"parent",u"ref",b"ref",u"spec",b"spec"]) -> None: ...

class SchemaRecord(google___protobuf___message___Message):
    DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...
    digest = ... # type: typing___Text
    blob = ... # type: builtin___bytes

    def __init__(self,
        *,
        digest : typing___Optional[typing___Text] = None,
        blob : typing___Optional[builtin___bytes] = None,
        ) -> None: ...
    @classmethod
    def FromString(cls, s: builtin___bytes) -> SchemaRecord: ...
    def MergeFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    def CopyFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    if sys.version_info >= (3,):
        def ClearField(self, field_name: typing_extensions___Literal[u"blob",u"digest"]) -> None: ...
    else:
        def ClearField(self, field_name: typing_extensions___Literal[u"blob",b"blob",u"digest",b"digest"]) -> None: ...

class PingRequest(google___protobuf___message___Message):
    DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...

    def __init__(self,
        ) -> None: ...
    @classmethod
    def FromString(cls, s: builtin___bytes) -> PingRequest: ...
    def MergeFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    def CopyFrom(self, other_msg: google___protobuf___message___Message) -> None: ...

class PingReply(google___protobuf___message___Message):
    DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...
    result = ... # type: typing___Text

    def __init__(self,
        *,
        result : typing___Optional[typing___Text] = None,
        ) -> None: ...
    @classmethod
    def FromString(cls, s: builtin___bytes) -> PingReply: ...
    def MergeFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    def CopyFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    if sys.version_info >= (3,):
        def ClearField(self, field_name: typing_extensions___Literal[u"result"]) -> None: ...
    else:
        def ClearField(self, field_name: typing_extensions___Literal[u"result",b"result"]) -> None: ...

class GetClientConfigRequest(google___protobuf___message___Message):
    DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...

    def __init__(self,
        ) -> None: ...
    @classmethod
    def FromString(cls, s: builtin___bytes) -> GetClientConfigRequest: ...
    def MergeFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    def CopyFrom(self, other_msg: google___protobuf___message___Message) -> None: ...

class GetClientConfigReply(google___protobuf___message___Message):
    DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...
    class ConfigEntry(google___protobuf___message___Message):
        DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...
        key = ... # type: typing___Text
        value = ... # type: typing___Text

        def __init__(self,
            *,
            key : typing___Optional[typing___Text] = None,
            value : typing___Optional[typing___Text] = None,
            ) -> None: ...
        @classmethod
        def FromString(cls, s: builtin___bytes) -> GetClientConfigReply.ConfigEntry: ...
        def MergeFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
        def CopyFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
        if sys.version_info >= (3,):
            def ClearField(self, field_name: typing_extensions___Literal[u"key",u"value"]) -> None: ...
        else:
            def ClearField(self, field_name: typing_extensions___Literal[u"key",b"key",u"value",b"value"]) -> None: ...


    @property
    def config(self) -> typing___MutableMapping[typing___Text, typing___Text]: ...

    @property
    def error(self) -> ErrorProto: ...

    def __init__(self,
        *,
        config : typing___Optional[typing___Mapping[typing___Text, typing___Text]] = None,
        error : typing___Optional[ErrorProto] = None,
        ) -> None: ...
    @classmethod
    def FromString(cls, s: builtin___bytes) -> GetClientConfigReply: ...
    def MergeFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    def CopyFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    if sys.version_info >= (3,):
        def HasField(self, field_name: typing_extensions___Literal[u"error"]) -> builtin___bool: ...
        def ClearField(self, field_name: typing_extensions___Literal[u"config",u"error"]) -> None: ...
    else:
        def HasField(self, field_name: typing_extensions___Literal[u"error",b"error"]) -> builtin___bool: ...
        def ClearField(self, field_name: typing_extensions___Literal[u"config",b"config",u"error",b"error"]) -> None: ...

class FetchBranchRecordRequest(google___protobuf___message___Message):
    DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...

    @property
    def rec(self) -> BranchRecord: ...

    def __init__(self,
        *,
        rec : typing___Optional[BranchRecord] = None,
        ) -> None: ...
    @classmethod
    def FromString(cls, s: builtin___bytes) -> FetchBranchRecordRequest: ...
    def MergeFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    def CopyFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    if sys.version_info >= (3,):
        def HasField(self, field_name: typing_extensions___Literal[u"rec"]) -> builtin___bool: ...
        def ClearField(self, field_name: typing_extensions___Literal[u"rec"]) -> None: ...
    else:
        def HasField(self, field_name: typing_extensions___Literal[u"rec",b"rec"]) -> builtin___bool: ...
        def ClearField(self, field_name: typing_extensions___Literal[u"rec",b"rec"]) -> None: ...

class FetchBranchRecordReply(google___protobuf___message___Message):
    DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...

    @property
    def rec(self) -> BranchRecord: ...

    @property
    def error(self) -> ErrorProto: ...

    def __init__(self,
        *,
        rec : typing___Optional[BranchRecord] = None,
        error : typing___Optional[ErrorProto] = None,
        ) -> None: ...
    @classmethod
    def FromString(cls, s: builtin___bytes) -> FetchBranchRecordReply: ...
    def MergeFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    def CopyFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    if sys.version_info >= (3,):
        def HasField(self, field_name: typing_extensions___Literal[u"error",u"rec"]) -> builtin___bool: ...
        def ClearField(self, field_name: typing_extensions___Literal[u"error",u"rec"]) -> None: ...
    else:
        def HasField(self, field_name: typing_extensions___Literal[u"error",b"error",u"rec",b"rec"]) -> builtin___bool: ...
        def ClearField(self, field_name: typing_extensions___Literal[u"error",b"error",u"rec",b"rec"]) -> None: ...

class FetchDataRequest(google___protobuf___message___Message):
    DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...
    raw_data = ... # type: builtin___bytes
    comp_nbytes = ... # type: builtin___int
    uncomp_nbytes = ... # type: builtin___int

    @property
    def error(self) -> ErrorProto: ...

    def __init__(self,
        *,
        raw_data : typing___Optional[builtin___bytes] = None,
        comp_nbytes : typing___Optional[builtin___int] = None,
        uncomp_nbytes : typing___Optional[builtin___int] = None,
        error : typing___Optional[ErrorProto] = None,
        ) -> None: ...
    @classmethod
    def FromString(cls, s: builtin___bytes) -> FetchDataRequest: ...
    def MergeFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    def CopyFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    if sys.version_info >= (3,):
        def HasField(self, field_name: typing_extensions___Literal[u"error"]) -> builtin___bool: ...
        def ClearField(self, field_name: typing_extensions___Literal[u"comp_nbytes",u"error",u"raw_data",u"uncomp_nbytes"]) -> None: ...
    else:
        def HasField(self, field_name: typing_extensions___Literal[u"error",b"error"]) -> builtin___bool: ...
        def ClearField(self, field_name: typing_extensions___Literal[u"comp_nbytes",b"comp_nbytes",u"error",b"error",u"raw_data",b"raw_data",u"uncomp_nbytes",b"uncomp_nbytes"]) -> None: ...

class FetchDataReply(google___protobuf___message___Message):
    DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...
    raw_data = ... # type: builtin___bytes
    comp_nbytes = ... # type: builtin___int
    uncomp_nbytes = ... # type: builtin___int

    @property
    def error(self) -> ErrorProto: ...

    def __init__(self,
        *,
        raw_data : typing___Optional[builtin___bytes] = None,
        comp_nbytes : typing___Optional[builtin___int] = None,
        uncomp_nbytes : typing___Optional[builtin___int] = None,
        error : typing___Optional[ErrorProto] = None,
        ) -> None: ...
    @classmethod
    def FromString(cls, s: builtin___bytes) -> FetchDataReply: ...
    def MergeFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    def CopyFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    if sys.version_info >= (3,):
        def HasField(self, field_name: typing_extensions___Literal[u"error"]) -> builtin___bool: ...
        def ClearField(self, field_name: typing_extensions___Literal[u"comp_nbytes",u"error",u"raw_data",u"uncomp_nbytes"]) -> None: ...
    else:
        def HasField(self, field_name: typing_extensions___Literal[u"error",b"error"]) -> builtin___bool: ...
        def ClearField(self, field_name: typing_extensions___Literal[u"comp_nbytes",b"comp_nbytes",u"error",b"error",u"raw_data",b"raw_data",u"uncomp_nbytes",b"uncomp_nbytes"]) -> None: ...

class FetchCommitRequest(google___protobuf___message___Message):
    DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...
    commit = ... # type: typing___Text

    def __init__(self,
        *,
        commit : typing___Optional[typing___Text] = None,
        ) -> None: ...
    @classmethod
    def FromString(cls, s: builtin___bytes) -> FetchCommitRequest: ...
    def MergeFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    def CopyFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    if sys.version_info >= (3,):
        def ClearField(self, field_name: typing_extensions___Literal[u"commit"]) -> None: ...
    else:
        def ClearField(self, field_name: typing_extensions___Literal[u"commit",b"commit"]) -> None: ...

class FetchCommitReply(google___protobuf___message___Message):
    DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...
    commit = ... # type: typing___Text
    total_byte_size = ... # type: builtin___int

    @property
    def record(self) -> CommitRecord: ...

    @property
    def error(self) -> ErrorProto: ...

    def __init__(self,
        *,
        commit : typing___Optional[typing___Text] = None,
        total_byte_size : typing___Optional[builtin___int] = None,
        record : typing___Optional[CommitRecord] = None,
        error : typing___Optional[ErrorProto] = None,
        ) -> None: ...
    @classmethod
    def FromString(cls, s: builtin___bytes) -> FetchCommitReply: ...
    def MergeFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    def CopyFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    if sys.version_info >= (3,):
        def HasField(self, field_name: typing_extensions___Literal[u"error",u"record"]) -> builtin___bool: ...
        def ClearField(self, field_name: typing_extensions___Literal[u"commit",u"error",u"record",u"total_byte_size"]) -> None: ...
    else:
        def HasField(self, field_name: typing_extensions___Literal[u"error",b"error",u"record",b"record"]) -> builtin___bool: ...
        def ClearField(self, field_name: typing_extensions___Literal[u"commit",b"commit",u"error",b"error",u"record",b"record",u"total_byte_size",b"total_byte_size"]) -> None: ...

class FetchSchemaRequest(google___protobuf___message___Message):
    DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...

    @property
    def rec(self) -> SchemaRecord: ...

    def __init__(self,
        *,
        rec : typing___Optional[SchemaRecord] = None,
        ) -> None: ...
    @classmethod
    def FromString(cls, s: builtin___bytes) -> FetchSchemaRequest: ...
    def MergeFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    def CopyFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    if sys.version_info >= (3,):
        def HasField(self, field_name: typing_extensions___Literal[u"rec"]) -> builtin___bool: ...
        def ClearField(self, field_name: typing_extensions___Literal[u"rec"]) -> None: ...
    else:
        def HasField(self, field_name: typing_extensions___Literal[u"rec",b"rec"]) -> builtin___bool: ...
        def ClearField(self, field_name: typing_extensions___Literal[u"rec",b"rec"]) -> None: ...

class FetchSchemaReply(google___protobuf___message___Message):
    DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...

    @property
    def rec(self) -> SchemaRecord: ...

    @property
    def error(self) -> ErrorProto: ...

    def __init__(self,
        *,
        rec : typing___Optional[SchemaRecord] = None,
        error : typing___Optional[ErrorProto] = None,
        ) -> None: ...
    @classmethod
    def FromString(cls, s: builtin___bytes) -> FetchSchemaReply: ...
    def MergeFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    def CopyFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    if sys.version_info >= (3,):
        def HasField(self, field_name: typing_extensions___Literal[u"error",u"rec"]) -> builtin___bool: ...
        def ClearField(self, field_name: typing_extensions___Literal[u"error",u"rec"]) -> None: ...
    else:
        def HasField(self, field_name: typing_extensions___Literal[u"error",b"error",u"rec",b"rec"]) -> builtin___bool: ...
        def ClearField(self, field_name: typing_extensions___Literal[u"error",b"error",u"rec",b"rec"]) -> None: ...

class PushBranchRecordRequest(google___protobuf___message___Message):
    DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...

    @property
    def rec(self) -> BranchRecord: ...

    def __init__(self,
        *,
        rec : typing___Optional[BranchRecord] = None,
        ) -> None: ...
    @classmethod
    def FromString(cls, s: builtin___bytes) -> PushBranchRecordRequest: ...
    def MergeFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    def CopyFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    if sys.version_info >= (3,):
        def HasField(self, field_name: typing_extensions___Literal[u"rec"]) -> builtin___bool: ...
        def ClearField(self, field_name: typing_extensions___Literal[u"rec"]) -> None: ...
    else:
        def HasField(self, field_name: typing_extensions___Literal[u"rec",b"rec"]) -> builtin___bool: ...
        def ClearField(self, field_name: typing_extensions___Literal[u"rec",b"rec"]) -> None: ...

class PushBranchRecordReply(google___protobuf___message___Message):
    DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...

    @property
    def error(self) -> ErrorProto: ...

    def __init__(self,
        *,
        error : typing___Optional[ErrorProto] = None,
        ) -> None: ...
    @classmethod
    def FromString(cls, s: builtin___bytes) -> PushBranchRecordReply: ...
    def MergeFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    def CopyFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    if sys.version_info >= (3,):
        def HasField(self, field_name: typing_extensions___Literal[u"error"]) -> builtin___bool: ...
        def ClearField(self, field_name: typing_extensions___Literal[u"error"]) -> None: ...
    else:
        def HasField(self, field_name: typing_extensions___Literal[u"error",b"error"]) -> builtin___bool: ...
        def ClearField(self, field_name: typing_extensions___Literal[u"error",b"error"]) -> None: ...

class PushDataRequest(google___protobuf___message___Message):
    DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...
    raw_data = ... # type: builtin___bytes
    comp_nbytes = ... # type: builtin___int
    uncomp_nbytes = ... # type: builtin___int

    @property
    def error(self) -> ErrorProto: ...

    def __init__(self,
        *,
        raw_data : typing___Optional[builtin___bytes] = None,
        comp_nbytes : typing___Optional[builtin___int] = None,
        uncomp_nbytes : typing___Optional[builtin___int] = None,
        error : typing___Optional[ErrorProto] = None,
        ) -> None: ...
    @classmethod
    def FromString(cls, s: builtin___bytes) -> PushDataRequest: ...
    def MergeFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    def CopyFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    if sys.version_info >= (3,):
        def HasField(self, field_name: typing_extensions___Literal[u"error"]) -> builtin___bool: ...
        def ClearField(self, field_name: typing_extensions___Literal[u"comp_nbytes",u"error",u"raw_data",u"uncomp_nbytes"]) -> None: ...
    else:
        def HasField(self, field_name: typing_extensions___Literal[u"error",b"error"]) -> builtin___bool: ...
        def ClearField(self, field_name: typing_extensions___Literal[u"comp_nbytes",b"comp_nbytes",u"error",b"error",u"raw_data",b"raw_data",u"uncomp_nbytes",b"uncomp_nbytes"]) -> None: ...

class PushDataReply(google___protobuf___message___Message):
    DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...

    @property
    def error(self) -> ErrorProto: ...

    def __init__(self,
        *,
        error : typing___Optional[ErrorProto] = None,
        ) -> None: ...
    @classmethod
    def FromString(cls, s: builtin___bytes) -> PushDataReply: ...
    def MergeFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    def CopyFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    if sys.version_info >= (3,):
        def HasField(self, field_name: typing_extensions___Literal[u"error"]) -> builtin___bool: ...
        def ClearField(self, field_name: typing_extensions___Literal[u"error"]) -> None: ...
    else:
        def HasField(self, field_name: typing_extensions___Literal[u"error",b"error"]) -> builtin___bool: ...
        def ClearField(self, field_name: typing_extensions___Literal[u"error",b"error"]) -> None: ...

class PushCommitRequest(google___protobuf___message___Message):
    DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...
    commit = ... # type: typing___Text
    total_byte_size = ... # type: builtin___int

    @property
    def record(self) -> CommitRecord: ...

    def __init__(self,
        *,
        commit : typing___Optional[typing___Text] = None,
        total_byte_size : typing___Optional[builtin___int] = None,
        record : typing___Optional[CommitRecord] = None,
        ) -> None: ...
    @classmethod
    def FromString(cls, s: builtin___bytes) -> PushCommitRequest: ...
    def MergeFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    def CopyFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    if sys.version_info >= (3,):
        def HasField(self, field_name: typing_extensions___Literal[u"record"]) -> builtin___bool: ...
        def ClearField(self, field_name: typing_extensions___Literal[u"commit",u"record",u"total_byte_size"]) -> None: ...
    else:
        def HasField(self, field_name: typing_extensions___Literal[u"record",b"record"]) -> builtin___bool: ...
        def ClearField(self, field_name: typing_extensions___Literal[u"commit",b"commit",u"record",b"record",u"total_byte_size",b"total_byte_size"]) -> None: ...

class PushCommitReply(google___protobuf___message___Message):
    DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...

    @property
    def error(self) -> ErrorProto: ...

    def __init__(self,
        *,
        error : typing___Optional[ErrorProto] = None,
        ) -> None: ...
    @classmethod
    def FromString(cls, s: builtin___bytes) -> PushCommitReply: ...
    def MergeFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    def CopyFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    if sys.version_info >= (3,):
        def HasField(self, field_name: typing_extensions___Literal[u"error"]) -> builtin___bool: ...
        def ClearField(self, field_name: typing_extensions___Literal[u"error"]) -> None: ...
    else:
        def HasField(self, field_name: typing_extensions___Literal[u"error",b"error"]) -> builtin___bool: ...
        def ClearField(self, field_name: typing_extensions___Literal[u"error",b"error"]) -> None: ...

class PushSchemaRequest(google___protobuf___message___Message):
    DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...

    @property
    def rec(self) -> SchemaRecord: ...

    def __init__(self,
        *,
        rec : typing___Optional[SchemaRecord] = None,
        ) -> None: ...
    @classmethod
    def FromString(cls, s: builtin___bytes) -> PushSchemaRequest: ...
    def MergeFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    def CopyFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    if sys.version_info >= (3,):
        def HasField(self, field_name: typing_extensions___Literal[u"rec"]) -> builtin___bool: ...
        def ClearField(self, field_name: typing_extensions___Literal[u"rec"]) -> None: ...
    else:
        def HasField(self, field_name: typing_extensions___Literal[u"rec",b"rec"]) -> builtin___bool: ...
        def ClearField(self, field_name: typing_extensions___Literal[u"rec",b"rec"]) -> None: ...

class PushSchemaReply(google___protobuf___message___Message):
    DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...

    @property
    def error(self) -> ErrorProto: ...

    def __init__(self,
        *,
        error : typing___Optional[ErrorProto] = None,
        ) -> None: ...
    @classmethod
    def FromString(cls, s: builtin___bytes) -> PushSchemaReply: ...
    def MergeFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    def CopyFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    if sys.version_info >= (3,):
        def HasField(self, field_name: typing_extensions___Literal[u"error"]) -> builtin___bool: ...
        def ClearField(self, field_name: typing_extensions___Literal[u"error"]) -> None: ...
    else:
        def HasField(self, field_name: typing_extensions___Literal[u"error",b"error"]) -> builtin___bool: ...
        def ClearField(self, field_name: typing_extensions___Literal[u"error",b"error"]) -> None: ...

class FindMissingCommitsRequest(google___protobuf___message___Message):
    DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...
    commits = ... # type: google___protobuf___internal___containers___RepeatedScalarFieldContainer[typing___Text]

    @property
    def branch(self) -> BranchRecord: ...

    def __init__(self,
        *,
        commits : typing___Optional[typing___Iterable[typing___Text]] = None,
        branch : typing___Optional[BranchRecord] = None,
        ) -> None: ...
    @classmethod
    def FromString(cls, s: builtin___bytes) -> FindMissingCommitsRequest: ...
    def MergeFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    def CopyFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    if sys.version_info >= (3,):
        def HasField(self, field_name: typing_extensions___Literal[u"branch"]) -> builtin___bool: ...
        def ClearField(self, field_name: typing_extensions___Literal[u"branch",u"commits"]) -> None: ...
    else:
        def HasField(self, field_name: typing_extensions___Literal[u"branch",b"branch"]) -> builtin___bool: ...
        def ClearField(self, field_name: typing_extensions___Literal[u"branch",b"branch",u"commits",b"commits"]) -> None: ...

class FindMissingCommitsReply(google___protobuf___message___Message):
    DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...
    commits = ... # type: google___protobuf___internal___containers___RepeatedScalarFieldContainer[typing___Text]

    @property
    def branch(self) -> BranchRecord: ...

    @property
    def error(self) -> ErrorProto: ...

    def __init__(self,
        *,
        commits : typing___Optional[typing___Iterable[typing___Text]] = None,
        branch : typing___Optional[BranchRecord] = None,
        error : typing___Optional[ErrorProto] = None,
        ) -> None: ...
    @classmethod
    def FromString(cls, s: builtin___bytes) -> FindMissingCommitsReply: ...
    def MergeFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    def CopyFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    if sys.version_info >= (3,):
        def HasField(self, field_name: typing_extensions___Literal[u"branch",u"error"]) -> builtin___bool: ...
        def ClearField(self, field_name: typing_extensions___Literal[u"branch",u"commits",u"error"]) -> None: ...
    else:
        def HasField(self, field_name: typing_extensions___Literal[u"branch",b"branch",u"error",b"error"]) -> builtin___bool: ...
        def ClearField(self, field_name: typing_extensions___Literal[u"branch",b"branch",u"commits",b"commits",u"error",b"error"]) -> None: ...

class FindMissingHashRecordsRequest(google___protobuf___message___Message):
    DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...
    commit = ... # type: typing___Text
    hashs = ... # type: builtin___bytes
    total_byte_size = ... # type: builtin___int

    def __init__(self,
        *,
        commit : typing___Optional[typing___Text] = None,
        hashs : typing___Optional[builtin___bytes] = None,
        total_byte_size : typing___Optional[builtin___int] = None,
        ) -> None: ...
    @classmethod
    def FromString(cls, s: builtin___bytes) -> FindMissingHashRecordsRequest: ...
    def MergeFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    def CopyFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    if sys.version_info >= (3,):
        def ClearField(self, field_name: typing_extensions___Literal[u"commit",u"hashs",u"total_byte_size"]) -> None: ...
    else:
        def ClearField(self, field_name: typing_extensions___Literal[u"commit",b"commit",u"hashs",b"hashs",u"total_byte_size",b"total_byte_size"]) -> None: ...

class FindMissingHashRecordsReply(google___protobuf___message___Message):
    DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...
    commit = ... # type: typing___Text
    hashs = ... # type: builtin___bytes
    total_byte_size = ... # type: builtin___int

    @property
    def error(self) -> ErrorProto: ...

    def __init__(self,
        *,
        commit : typing___Optional[typing___Text] = None,
        hashs : typing___Optional[builtin___bytes] = None,
        total_byte_size : typing___Optional[builtin___int] = None,
        error : typing___Optional[ErrorProto] = None,
        ) -> None: ...
    @classmethod
    def FromString(cls, s: builtin___bytes) -> FindMissingHashRecordsReply: ...
    def MergeFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    def CopyFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    if sys.version_info >= (3,):
        def HasField(self, field_name: typing_extensions___Literal[u"error"]) -> builtin___bool: ...
        def ClearField(self, field_name: typing_extensions___Literal[u"commit",u"error",u"hashs",u"total_byte_size"]) -> None: ...
    else:
        def HasField(self, field_name: typing_extensions___Literal[u"error",b"error"]) -> builtin___bool: ...
        def ClearField(self, field_name: typing_extensions___Literal[u"commit",b"commit",u"error",b"error",u"hashs",b"hashs",u"total_byte_size",b"total_byte_size"]) -> None: ...

class FindMissingSchemasRequest(google___protobuf___message___Message):
    DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...
    commit = ... # type: typing___Text
    schema_digests = ... # type: google___protobuf___internal___containers___RepeatedScalarFieldContainer[typing___Text]

    def __init__(self,
        *,
        commit : typing___Optional[typing___Text] = None,
        schema_digests : typing___Optional[typing___Iterable[typing___Text]] = None,
        ) -> None: ...
    @classmethod
    def FromString(cls, s: builtin___bytes) -> FindMissingSchemasRequest: ...
    def MergeFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    def CopyFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    if sys.version_info >= (3,):
        def ClearField(self, field_name: typing_extensions___Literal[u"commit",u"schema_digests"]) -> None: ...
    else:
        def ClearField(self, field_name: typing_extensions___Literal[u"commit",b"commit",u"schema_digests",b"schema_digests"]) -> None: ...

class FindMissingSchemasReply(google___protobuf___message___Message):
    DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...
    commit = ... # type: typing___Text
    schema_digests = ... # type: google___protobuf___internal___containers___RepeatedScalarFieldContainer[typing___Text]

    @property
    def error(self) -> ErrorProto: ...

    def __init__(self,
        *,
        commit : typing___Optional[typing___Text] = None,
        schema_digests : typing___Optional[typing___Iterable[typing___Text]] = None,
        error : typing___Optional[ErrorProto] = None,
        ) -> None: ...
    @classmethod
    def FromString(cls, s: builtin___bytes) -> FindMissingSchemasReply: ...
    def MergeFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    def CopyFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    if sys.version_info >= (3,):
        def HasField(self, field_name: typing_extensions___Literal[u"error"]) -> builtin___bool: ...
        def ClearField(self, field_name: typing_extensions___Literal[u"commit",u"error",u"schema_digests"]) -> None: ...
    else:
        def HasField(self, field_name: typing_extensions___Literal[u"error",b"error"]) -> builtin___bool: ...
        def ClearField(self, field_name: typing_extensions___Literal[u"commit",b"commit",u"error",b"error",u"schema_digests",b"schema_digests"]) -> None: ...
